theory AIRSyntax
  imports Main Word_Lib.Word_Lib_Sumo MachineState
begin


text \<open>Helpful lemma for the map types\<close>

lemma dom_ran: \<open>y \<in> dom x \<Longrightarrow> the (x y) \<in> ran x\<close>
  by (simp add: domD ranI)

text \<open>Predicate indicating that the state is currently halting (end of program)\<close>

definition \<open>halting \<rho>' \<Pi>' \<longleftrightarrow> \<rho>' \<notin> dom \<Pi>'\<close>

definition \<open>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<longleftrightarrow> halting (\<rho> s) (\<Pi> s)\<close>

text \<open>Predicate indicating that the state is currently speculating\<close>

definition speculating :: \<open>MachineStatePredicate\<close> where
  \<open>speculating s \<longleftrightarrow> n s > 0\<close>

text \<open>Predicate indicating that the state is terminating\<close>

definition
  terminating :: \<open>MachineStatePredicate\<close>
where
  \<open>terminating s \<longleftrightarrow> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<and> \<not>speculating s\<close>



subsection \<open>Speculative branch resolution\<close>

text \<open>A branch can be resolved at any point (non-deterministic) given that we are currently 
      speculating (n > 0). If we are currently halting (end of program) then the only action we can 
      take is to resolve until n = 0.\<close>

definition \<open>resolve p\<^sub>n p\<^sub>\<beta> p\<^sub>p\<^sub>c \<longleftrightarrow> (\<some> x. x \<in> UNIV)\<close>

subsection \<open>Speculative branch miss-predictions\<close>

text \<open>The branch predictor may mispredict the outcome of a branch expression, triggering 
      speculative execution. This is non-deterministic, making no assumptions as to the state of the 
      branch predictor and it's likely prediction.\<close>

definition \<open>mispred p\<^sub>n p\<^sub>\<beta> p\<^sub>p\<^sub>c \<longleftrightarrow> pred p\<^sub>\<beta> \<and> (\<some> x. x \<in> UNIV)\<close>

lemma \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred: \<open>\<not>\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<Longrightarrow> \<not> mispred (n s) (\<beta> s) (pc s)\<close>
  by (simp add: mispred_def)

subsection \<open>Updating the state of the branch predictor\<close>

text \<open>The branch predictor is updated by appending the speculation level and program address. 
      A mismatch in traces between two branch predictors indicates the divergence of paths.\<close>

definition update :: \<open>SpeculationLevel \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> bool \<Rightarrow> BranchPredictorState \<Rightarrow> BranchPredictorState\<close> where
  \<open>update p\<^sub>n p\<^sub>\<rho> e p\<^sub>\<beta> = p\<^sub>\<beta> \<lparr>trace := (trace p\<^sub>\<beta>) @ [(p\<^sub>n, p\<^sub>\<rho>, e)]\<rparr>\<close>

lemma \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const: \<open>\<beta>'' = update n' ptr' eval' \<beta>' \<Longrightarrow> pred \<beta>'' = pred \<beta>'\<close>
  using update_def by auto

text \<open>Increment the program address for the current state and store as a Addr$_\<Pi>$. This is 
      always greater unless the program address is at the max address.\<close>

definition \<open>next_\<rho> p\<^sub>\<rho> \<equiv> (p\<^sub>\<rho> + 1)\<close>

lemma next_\<rho>_greater: \<open>\<rho> s < (next_\<rho> (\<rho> s))\<close>
  by (auto simp add: next_\<rho>_def)

lemma next_\<rho>_gteq: \<open>\<rho> s \<le> (next_\<rho> (\<rho> s))\<close>
  using less_or_eq_imp_le next_\<rho>_greater by presburger


subsection \<open>Well Formedness\<close>

text \<open>The Speculation level is well formed if it is greater than 0 (implied by nat type)\<close>

definition \<open>wfn s \<equiv> \<not>(\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s) \<longrightarrow> n s = 0\<close>

text \<open>The register state is well formed if for all speculation levels it is defined.\<close>

definition \<open>wf\<Delta> s \<equiv> (\<forall>n'. n' \<le> (n s) \<longrightarrow> \<Delta> s n' \<noteq> None) \<and> (\<forall>n'. n' > (n s) \<longrightarrow> \<Delta> s n' = None)\<close>

text \<open>The data memory is well formed if for all speculation levels it is defined.\<close>

definition \<open>wf\<mu> s \<equiv> (\<forall>n'. n' \<le> (n s) \<longrightarrow> \<mu> s n' \<noteq> None) \<and> (\<forall>n'. n' > (n s) \<longrightarrow> \<mu> s n' = None)\<close>

text \<open>The program counter is well formed if for all speculation levels it is defined and the address
      it points to is valid.\<close>

definition \<open>wfpc s \<equiv> (\<forall>n'. n' \<le> (n s) \<longrightarrow> (pc s n' \<noteq> None)) \<and> (\<forall>n'. n' > (n s) \<longrightarrow> pc s n' = None)\<close>

text \<open>The program trace is well formed if it's program address is in the domain of the program 
      memory.\<close>

definition \<open>wf\<omega> s \<equiv> (\<forall>a\<^sub>\<Pi> a\<^sub>\<mu>. (a\<^sub>\<Pi>, a\<^sub>\<mu>) \<in> set (\<omega> s) \<longrightarrow> a\<^sub>\<Pi> \<in> \<Pi>\<^sub>\<rho> s)\<close>

text \<open>The branch predictor is well formed if it's program address is in the domain of the program 
      memory and the instruction at that address equals the logged instruction.\<close>

definition \<open>wf\<beta> s \<equiv> (\<forall>\<beta>\<^sub>\<omega>\<^sub>n \<beta>\<^sub>\<omega>\<^sub>\<rho> \<beta>\<^sub>\<omega>\<^sub>m\<^sub>i\<^sub>s\<^sub>p\<^sub>r\<^sub>e\<^sub>d . (\<beta>\<^sub>\<omega>\<^sub>n, \<beta>\<^sub>\<omega>\<^sub>\<rho>, \<beta>\<^sub>\<omega>\<^sub>m\<^sub>i\<^sub>s\<^sub>p\<^sub>r\<^sub>e\<^sub>d) \<in> set (\<beta>\<^sub>\<omega> s) 
  \<longrightarrow> (\<not>(\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s) \<longrightarrow> \<not>\<beta>\<^sub>\<omega>\<^sub>m\<^sub>i\<^sub>s\<^sub>p\<^sub>r\<^sub>e\<^sub>d \<and> \<beta>\<^sub>\<omega>\<^sub>n = 0)
)\<close>

lemma wf\<beta>_update:
  assumes \<open>wf\<beta> s\<close>
      and \<open>wfpc s\<close>
      and \<open>wfn s\<close>
      and \<open>\<not> \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<longrightarrow> \<not> miss\<close>
      and \<open>s' = s\<lparr> \<beta> := update (n s) (\<rho>\<^sub>p\<^sub>t\<^sub>r s) miss (\<beta> s)\<rparr>\<close>
    shows \<open>wf\<beta> s'\<close>
  using assms by (auto simp add: wf\<beta>_def update_def wfpc_def wfn_def Let_def) 

text \<open>The trusted instruction set is well formed if all addresses are in the program memory.\<close>

definition \<open>wf\<T>\<^sub>\<rho> s \<equiv> (\<T>\<^sub>\<rho> s \<subseteq> \<Pi>\<^sub>\<rho> s)\<close>

text \<open>The entrypoint is well formed if it is in the trusted set.\<close>

definition \<open>wf\<E>\<P> s \<equiv> \<E>\<P> s \<in> \<T>\<^sub>\<rho> s\<close>

text \<open>The adversary readable memory addresses are well formed if they are defined in the program 
      memory for all speculation levels and are not in the secret set (otherwise trivial).\<close>

definition \<open>wf\<U>rd s \<equiv> (\<forall>a. a \<in> \<U>rd s \<longrightarrow> a \<notin> \<S>\<^sub>\<T> s)\<close>

text \<open>The adversary writable memory addresses are well formed if they are defined in the program 
      memory for all speculation levels and are not in the secret set (otherwise trivial).\<close>

definition \<open>wf\<U>wr s \<equiv> (\<forall>a. a \<in> \<U>wr s \<longrightarrow> a \<notin> \<S>\<^sub>\<T> s)\<close>

text \<open>The wfs(s) predicate indicates that the state s is well-formed. Well-formedness is invariant
for any program.\<close>

definition \<open>wfs s \<equiv> 
  wf\<Delta> s \<and>
  wf\<mu> s \<and>
  wfpc s \<and>
  wf\<omega> s \<and>
  wf\<beta> s \<and>
  wfn s \<and>
  wf\<T>\<^sub>\<rho> s \<and>
  wf\<E>\<P> s \<and>
  wf\<U>rd s \<and>
  wf\<U>wr s
\<close>

text \<open>If not halting then \<rho> in \<Pi>\<close>

lemma not_halting_\<rho>_in_\<Pi>: \<open>wfpc s \<Longrightarrow> \<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<Longrightarrow> \<rho> s \<in> \<Pi>\<^sub>\<rho> s\<close>
  by (simp add: halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def halting_def)

text \<open>If not speculating then two program counters speculative values should trivially be the same\<close>

lemma ns_pc_s_equal:
  assumes \<open>\<not>speculating s\<^sub>1\<close>
      and \<open>\<not>speculating s\<^sub>2\<close>
      and \<open>wfs s\<^sub>1\<close>
      and \<open>wfs s\<^sub>2\<close>
    shows \<open>\<forall>n. n > 0 \<longrightarrow> pc s\<^sub>1 n = pc s\<^sub>2 n\<close>
  using assms by (auto simp add: speculating_def wfs_def wfpc_def)

lemma n_pc_implies_pc_equal:
  assumes \<open>\<forall>n. pc s\<^sub>1 n = pc s\<^sub>2 n\<close>
    shows \<open>pc s\<^sub>1 = pc s\<^sub>2\<close>
  using assms by blast

lemma ns_pc_\<rho>_eq:
  assumes \<open>\<not>speculating s\<^sub>1\<close>
      and \<open>\<not>speculating s\<^sub>2\<close>
      and \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>1\<close>
      and \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>2\<close>
      and \<open>wfs s\<^sub>1\<close>
      and \<open>wfs s\<^sub>2\<close>
      and \<open>\<rho> s\<^sub>1 = \<rho> s\<^sub>2\<close>
    shows \<open>pc s\<^sub>1 = pc s\<^sub>2\<close>
proof -
  have \<open>pc s\<^sub>1 0 = pc s\<^sub>2 0\<close>
    using assms(1-7) apply (simp add: speculating_def halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def halting_def wfs_def wfpc_def)
    by (metis option.sel)

  then show ?thesis
    by (metis assms(1) assms(2) assms(5) assms(6) less_nat_zero_code n_pc_implies_pc_equal nat_neq_iff ns_pc_s_equal)
qed


lemma \<mu>\<^sub>u\<^sub>n\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r\<^sub>a\<^sub>i\<^sub>n\<^sub>e\<^sub>d_\<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d_not_all_equal:
  assumes \<open>\<mu>\<^sub>n s\<^sub>1 = \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t m u\<^sub>1\<close>
      and \<open>\<mu>\<^sub>n s\<^sub>2 = \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t m u\<^sub>2\<close>
      and \<open>\<mu>\<^sub>u\<^sub>n\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r\<^sub>a\<^sub>i\<^sub>n\<^sub>e\<^sub>d u\<^sub>1 u\<^sub>2 m\<close>
    shows \<open>\<exists>a. \<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s\<^sub>1 a \<noteq> \<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s\<^sub>2 a\<close>
  using assms apply (simp only: \<mu>\<^sub>u\<^sub>n\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r\<^sub>a\<^sub>i\<^sub>n\<^sub>e\<^sub>d_def)
  by (metis \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_a_not_in_map_simp)


(* TODO tidy these up and use MachineWord if possible *)
definition \<open>as_word b \<equiv> (if b then 0x1 else 0x0)\<close>

definition \<open>air_lt w1 w2 \<equiv> as_word (w1 < w2)\<close>
definition air_shiftl :: \<open>64 word \<Rightarrow> 64 word \<Rightarrow> 64 word\<close> where
\<open>air_shiftl w1 w2 \<equiv> w2 << unat w1\<close>



end
