theory MachineState
  imports AIR
begin

section \<open>MachineState\<close>


subsection \<open>Data Types\<close>

text \<open>The speculation level (n) refers to a natural number representing the level of speculation\<close>

type_synonym SpeculationLevel = nat

text \<open>The register and data memory states are maps of the speculation level and address/register to 
a Constant value. This models the state of registers at different speculation 
levels\<close>

type_synonym Registers = \<open>Reg \<rightharpoonup> Const\<close>
type_synonym RegisterState = \<open>SpeculationLevel \<rightharpoonup> Registers\<close>
type_synonym DataMemory = \<open>Addr\<^sub>\<mu> \<Rightarrow> Const\<close>
type_synonym DataMemoryState = \<open>SpeculationLevel \<rightharpoonup> DataMemory\<close>
                                                                    

text \<open>The program counter indexed by speculation level\<close>

type_synonym ProgramCounter = \<open>SpeculationLevel \<rightharpoonup> Addr\<^sub>\<Pi>\<close>

text \<open>Branch predictor state, "predict" indicates whether the branch predictor can make 
      mispredictions and "trace" describes a trace of the speculation level, addresses and executed 
      Instruction.\<close>

type_synonym BranchPredictor = \<open>(SpeculationLevel \<times> Addr\<^sub>\<Pi> \<times> bool) list\<close>

text \<open>Program Memory, a map of addresses to program Instructions.\<close>

type_synonym ProgramMemory = \<open>Addr\<^sub>\<Pi> \<rightharpoonup> Instr\<close>

text \<open>Trace of accessed program addresses and data addresses (which are optional).\<close>
                                       
type_synonym AddressTrace = \<open>(Addr\<^sub>\<Pi> \<times> Addr\<^sub>\<mu> option) list\<close>


subsection \<open>The machine state record\<close>

record MachineState =
  \<Pi> :: Prog
  \<Delta> :: RegisterState
  \<mu> :: DataMemoryState
  pc :: ProgramCounter
  \<omega> :: AddressTrace
  \<beta> :: BranchPredictor
  n :: SpeculationLevel
  \<T>\<^sub>\<rho> :: \<open>Addr\<^sub>\<Pi> set\<close>
  \<E>\<P> :: \<open>Addr\<^sub>\<Pi>\<close>
  \<S>\<^sub>\<T> :: \<open>Addr\<^sub>\<mu> set\<close>
  \<U>rd :: \<open>Addr\<^sub>\<mu> set\<close>
  \<U>wr :: \<open>Addr\<^sub>\<mu> set\<close>

text \<open>Define predicates and transformers for MachineState\<close>

type_synonym MachineStatePredicate = \<open>MachineState \<Rightarrow> bool\<close>
type_synonym MachineStateTransformer = \<open>MachineState \<Rightarrow> MachineState\<close>


subsection \<open>Machine state functions\<close>

text \<open>The set of program addresses in the program memory\<close>

abbreviation
  \<Pi>\<^sub>\<rho> :: \<open>MachineState \<Rightarrow> Addr\<^sub>\<Pi> set\<close> 
where
  \<open>\<Pi>\<^sub>\<rho> s \<equiv> dom (\<Pi> s)\<close>

text \<open>The set of instructions in the program memory\<close>

abbreviation
  \<Pi>\<^sub>\<iota> :: \<open>MachineState \<Rightarrow> Instr set\<close> 
where
  \<open>\<Pi>\<^sub>\<iota> s \<equiv> ran (\<Pi> s)\<close>

text \<open>Initialization for unConstrained data memory with map. These definitions and lemmas exist to 
      make solving examples concise.\<close>

definition 
  \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t :: \<open>(Addr\<^sub>\<mu> \<rightharpoonup> Const) \<Rightarrow> DataMemory \<Rightarrow> DataMemory\<close> 
where
  \<open>\<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t m u \<equiv> \<lambda>a. case m a of Some v \<Rightarrow> v | None \<Rightarrow> u a\<close>

lemma \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_a_in_map_simp: \<open>a \<in> dom m \<Longrightarrow> \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t m u a = the (m a)\<close>
  by (metis \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_def domD option.sel option.simps(5))

lemma \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_a_not_in_map_simp: \<open>a \<notin> dom m \<Longrightarrow> \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t m u a = u a\<close>
  by (simp add: \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_def domIff)

text \<open>Write to an unConstrained data memory with map\<close>

definition
  \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e :: \<open>Addr\<^sub>\<mu> \<Rightarrow> Const \<Rightarrow> DataMemory \<Rightarrow> DataMemory\<close> 
where
 \<open>\<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e a v u \<equiv> \<lambda>a'. if a' = a then v else u a'\<close>

lemma \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e_\<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_simp: \<open>\<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e a v (\<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t m u) = \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t (m(a \<mapsto> v)) u\<close>
  by (auto simp add: \<mu>\<^sub>n\<^sub>i\<^sub>n\<^sub>i\<^sub>t_def \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e_def)

text \<open>Memory at the current speculation level\<close>

abbreviation
  \<mu>\<^sub>n :: \<open>MachineState \<Rightarrow> DataMemory\<close>
where
  \<open>\<mu>\<^sub>n s \<equiv> the ((\<mu> s) (n s))\<close>

text \<open>UnConstrained memory read at the current speculation level for the given address\<close>

abbreviation
  \<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d :: \<open>MachineState \<Rightarrow> Addr\<^sub>\<mu> \<Rightarrow> Const\<close> 
where
  \<open>\<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s a \<equiv> (\<mu>\<^sub>n s) a\<close>

definition \<open>\<mu>\<^sub>u\<^sub>n\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r\<^sub>a\<^sub>i\<^sub>n\<^sub>e\<^sub>d u\<^sub>1 u\<^sub>2 m \<longleftrightarrow> (\<exists>a. a \<notin> dom m \<and> u\<^sub>1 a \<noteq> u\<^sub>2 a)\<close>

text \<open>Memory write at the current speculation level\<close> 

abbreviation
  \<mu>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e :: \<open>MachineState \<Rightarrow> Addr\<^sub>\<mu> \<Rightarrow> Const \<Rightarrow> DataMemoryState\<close> 
where
  \<open>\<mu>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s a v \<equiv> (\<mu> s)(n s \<mapsto> \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e a v (\<mu>\<^sub>n s))\<close>

text \<open>Copy the data memory at the current speculation level to the next speculation level\<close> 

abbreviation
  \<mu>\<^sub>s\<^sub>p\<^sub>e\<^sub>c :: \<open>MachineState \<Rightarrow> DataMemoryState\<close> 
where
  \<open>\<mu>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s \<equiv> (\<mu> s)(n s + 1 \<mapsto> \<mu>\<^sub>n s)\<close>

text \<open>Remove the data memory at the current speculation level\<close> 

abbreviation
  \<mu>\<^sub>r\<^sub>e\<^sub>s :: \<open>MachineState \<Rightarrow> DataMemoryState\<close> 
where
  \<open>\<mu>\<^sub>r\<^sub>e\<^sub>s s \<equiv> (\<mu> s)(n s := None)\<close>
 
text \<open>Registers at the current speculation level\<close>

abbreviation
  \<Delta>\<^sub>n :: \<open>MachineState \<Rightarrow> Registers\<close> 
where
  \<open>\<Delta>\<^sub>n s \<equiv> the (\<Delta> s (n s))\<close>

text \<open>UnConstrained memory read at the current speculation level for the given address\<close>

abbreviation
  \<Delta>\<^sub>r\<^sub>e\<^sub>a\<^sub>d :: \<open>MachineState \<Rightarrow> Reg \<Rightarrow> Const\<close> 
where
  \<open>\<Delta>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s r \<equiv> the (\<Delta>\<^sub>n s r)\<close>

text \<open>Memory write at the current speculation level\<close> 

abbreviation
  \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e :: \<open>MachineState \<Rightarrow> Reg \<Rightarrow> Const \<Rightarrow> RegisterState\<close> 
where
  \<open>\<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r v \<equiv> (\<Delta> s)(n s \<mapsto> \<Delta>\<^sub>n s(r \<mapsto> v))\<close>

text \<open>\<close>

abbreviation
  \<Delta>\<^sub>s\<^sub>p\<^sub>e\<^sub>c :: \<open>MachineState \<Rightarrow> RegisterState\<close> 
where
  \<open>\<Delta>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s \<equiv> (\<Delta> s)(n s + 1 \<mapsto> \<Delta>\<^sub>n s)\<close>

text \<open>\<close>

abbreviation
  \<Delta>\<^sub>r\<^sub>e\<^sub>s :: \<open>MachineState \<Rightarrow> RegisterState\<close>
where
  \<open>\<Delta>\<^sub>r\<^sub>e\<^sub>s s \<equiv> (\<Delta> s)(n s := None)\<close>

text \<open>Retrieve the program address for the current state\<close>

abbreviation 
  \<rho> :: \<open>MachineState \<Rightarrow> Addr\<^sub>\<Pi>\<close> 
where
  \<open>\<rho> s \<equiv> the (pc s (n s))\<close>

text \<open>Retrieve the Instruction for the current state\<close>

abbreviation 
  \<iota> :: \<open>MachineState \<Rightarrow> Instr\<close> 
where
  \<open>\<iota> s \<equiv> the (\<Pi> s (\<rho> s))\<close>

lemma \<iota>_some_Instr:
  \<open>(\<exists> reg exp. \<iota> s = RegisterUpdate reg exp) \<or>
   (\<exists> reg exp. \<iota> s = Load reg exp) \<or>
   (\<exists> exp1 exp2. \<iota> s = Store exp1 exp2) \<or>
   (\<exists> Addr\<^sub>\<Pi> exp. \<iota> s = Branch exp Addr\<^sub>\<Pi>) \<or>
   (\<exists> Addr\<^sub>\<Pi>. \<iota> s = Goto Addr\<^sub>\<Pi>) \<or>
   \<iota> s = SpecFence \<or>
   \<iota> s = NoOp\<close>
  apply auto
  by (meson Instr.exhaust)

end