theory AIR
  imports Main Word_Lib.Word_64
begin

section \<open>AIR (Assembly Intermediate Representation)\<close>

text \<open>Specific instruction set architectures (ISA) such as x86 and ARM would be complex to verify 
      and could hide the fundamental cause of transient execution vulnerabilities. Instead we lift
      ISAs into AIR (Assembly Intermediate Representation), an intermediary language who's syntax 
      is defined in this document.\<close>

text \<open>First we define\<close>

text \<open>Const is the set of all constants, realised as a 64 bit uninterpreted binary word.\<close>

type_synonym Const = word64

text \<open>Reg defines the set of all Registers. "V (nat)" refers to a virtual (or temporary) Register 
      which is assigned some numeric label. We make no assumptions on the number of virtual 
      Registers available in AIR.\<close>

datatype Reg = RAX | RBX | RCX | RDX | RDI | RSI | RSP | RBP | CF | ZF | V nat

text \<open>We let Addr$_\<Pi>$ and Addr$_\<mu>$ be the set of all program and memory addresses respectively. 
      Both of which being Const.\<close> 

type_synonym Addr\<^sub>\<Pi> = nat
type_synonym Addr\<^sub>\<mu> = Const

text \<open>Binary (\<lozenge>$_b$) and Unary (\<lozenge>$_u$) operators are modelled as functions across Const types.\<close>

type_synonym UnOp = \<open>Const \<Rightarrow> Const\<close>
type_synonym BinOp = \<open>Const \<Rightarrow> Const \<Rightarrow> Const\<close>

text \<open>An Expression in AIR can be one of four types:

Const - evaluates to the specified constant
Reg   - read the constant in the specified Register
UnOp  - evaluate Expression and apply unary operator to the constant result
BinOp - evaluate both Expressions and apply binary operator to the constant results

Evaluation of these Expressions is shown the Operational Semantics
\<close>

datatype Exp =
    Const Const
  | Reg Reg 
  | UnOp UnOp Exp
  | BinOp Exp BinOp Exp


subsection \<open>Instructions\<close>

text \<open>The instructions are standard AIR syntax with the addition of SpecFence.

RegisterUpdate - Update the given Register
Load           - Load the value from memory into the given Register
Store          - Store the value in memory
Branch         - If the Expression evaluates to true (1) then go to the program address
Goto           - Go to the program address
SpecFence      - Force the processor to resolve outstanding branches
NoOp           - No operation/instruction. Skip this instruction and proceed to next.

Evaluation of these instructions is shown the Operational Semantics
.\<close>

datatype Instr =
    RegisterUpdate Reg Exp
  | Load Reg Exp
  | Store Exp Exp
  | Branch Exp Addr\<^sub>\<Pi>
  | Goto Addr\<^sub>\<Pi>
  | SpecFence
  | NoOp

text \<open>Programs (Prog) are a map from program addresses (Addr$_\<Pi>$) to instructions Instr.\<close>

type_synonym Prog = \<open>Addr\<^sub>\<Pi> \<rightharpoonup> Instr\<close>


subsection \<open>BIL (Binary Intermediate Language) abbreviations\<close>

text \<open>AIR is often generated from BIL, the following abbreviations match BIL syntax to AIR for 
      unsigned 64 bit integers.\<close>

abbreviation register_update_alias :: \<open>Reg \<Rightarrow> Exp \<Rightarrow> Instr\<close> (\<open>_ := _\<close>) where
\<open>(r := e) \<equiv> RegisterUpdate r e\<close>

abbreviation load_alias :: \<open>Reg \<Rightarrow> Exp \<Rightarrow> Instr\<close> (\<open>_ := mem[_, el]:u64\<close>) where
\<open>(r := mem[e, el]:u64) \<equiv> Load r e\<close>

abbreviation store_alias :: \<open>Exp \<Rightarrow> Exp \<Rightarrow> Instr\<close> (\<open>mem := mem with [_, el]:u64 <- _\<close>) where
\<open>(mem := mem with [e\<^sub>1, el]:u64 <- e\<^sub>2) \<equiv> Store e\<^sub>1 e\<^sub>2\<close>

abbreviation branch_alias :: \<open>Exp \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> Instr\<close> (\<open>when _ goto _\<close>) where
\<open>(when e goto c) \<equiv> Branch e c\<close>

abbreviation goto_alias :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Instr\<close>  (\<open>goto _\<close>) where
\<open>(goto c) \<equiv> Goto c\<close>

abbreviation \<open>specfence \<equiv> SpecFence\<close>

abbreviation \<open>noop \<equiv> NoOp\<close>

end